# TODO: use retry library instead of custom code. Tenacity is one option.
import random


class RetryStrategy:
    statuses = {503, 524, 520, 504}

    @staticmethod
    def sleep_times():
        sleep_times = [1, 3, 9, 27]  # These are in seconds

        return [2 * random.random() * t for t in sleep_times]



import json
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Optional

import requests

from .constants import (
    DATASET_ID_KEY,
    METADATA_KEY,
    OVERWRITE_KEY,
    REFERENCE_ID_KEY,
)

if TYPE_CHECKING:
    from . import Connection


@dataclass  # pylint: disable=R0902
class Track:  # pylint: disable=R0902
    """A track is a class of objects (annotation or prediction) that forms a one-to-many relationship
    with objects, wherein an object is an instance of a track.

    Args:
        reference_id (str): A user-specified name of the track that describes the class of objects it represents.
        metadata: Arbitrary key/value dictionary of info to attach to this track.
    """

    _connection: "Connection"
    dataset_id: str
    reference_id: str
    metadata: Optional[dict] = None

    def __repr__(self):
        return f"Track(dataset_id='{self.dataset_id}', reference_id='{self.reference_id}', metadata={self.metadata})"

    def __eq__(self, other):
        return (
            (self.dataset_id == other.dataset_id)
            and (self.reference_id == other.reference_id)
            and (self.metadata == other.metadata)
        )

    @classmethod
    def from_json(cls, payload: dict, connection: "Connection"):
        """Instantiates track object from schematized JSON dict payload."""
        return cls(
            _connection=connection,
            reference_id=str(payload[REFERENCE_ID_KEY]),
            dataset_id=str(payload[DATASET_ID_KEY]),
            metadata=payload.get(METADATA_KEY, None),
        )

    def to_payload(self) -> dict:
        """Serializes track object to schematized JSON dict."""
        payload: Dict[str, Any] = {
            REFERENCE_ID_KEY: self.reference_id,
            DATASET_ID_KEY: self.dataset_id,
            METADATA_KEY: self.metadata,
        }

        return payload

    def to_json(self) -> str:
        """Serializes track object to schematized JSON string."""
        return json.dumps(self.to_payload(), allow_nan=False)

    def update(
        self,
        metadata: Optional[dict] = None,
        overwrite_metadata: bool = False,
    ) -> None:
        """
        Updates the Track's metadata.

        Parameters:
            metadata (Optional[dict]): An arbitrary dictionary of additional data about this track that can be stored
                and retrieved.
            overwrite_metadata (Optional[bool]): If metadata is provided and overwrite_metadata = True, then the track's
                entire metadata object will be overwritten. Otherwise, only the keys in metadata will be overwritten.
        """

        self._connection.make_request(
            payload={
                REFERENCE_ID_KEY: self.reference_id,
                METADATA_KEY: metadata,
                OVERWRITE_KEY: overwrite_metadata,
            },
            route=f"dataset/{self.dataset_id}/tracks",
            requests_command=requests.post,
        )
        self.metadata = (
            metadata
            if overwrite_metadata
            else (
                {**self.metadata, **metadata}
                if self.metadata is not None and metadata is not None
                else metadata
            )
        )



from collections import defaultdict
from enum import Enum
from typing import Any, Dict, List

from .async_job import AsyncJob


class CustomerJobTypes(str, Enum):
    UPLOAD_DATASET_ITEMS = "uploadDatasetItems"
    UPLOAD_PREDICTIONS = "uploadPredictions"
    UPLOAD_ANNOTATIONS = "uploadAnnotations"
    UPLOAD_LIDAR_SCENE = "uploadLidarScene"
    UPLOAD_VIDEO_SCENE = "uploadVideoScene"
    MODEL_INFERENCE_RUN = "modelInferenceRun"
    INDEXING_IMAGE = "indexingImage"
    INDEXING_OBJECT = "indexingObject"
    ANNOTATION_DELETION = "annotationDeletion"
    SEND_TO_LABELING = "sendToLabeling"
    INGEST_TASKS = "ingestTasks"
    CUSTOM_INDEXING = "customIndexing"
    EMBEDDING_DELETE = "embeddingDelete"
    TEST_EVALUATION = "testEvaluation"
    VALIDATE_METRICS = "modelMetrics"
    MODEL_RUN_COMMIT = "modelRunCommit"
    AUTOTAG_INFERENCE = "autotagInference"
    SLICE_QUERY = "sliceQuery"
    CLONE_DATASET = "cloneDataset"
    METADATA_UPDATE = "metadataUpdate"
    TRIGGER_EVALUATE = "triggerEvaluate"
    EXPORT_EMBEDDINGS = "exportEmbeddings"

    def __contains__(self, item):
        try:
            self(item)
        except ValueError:
            return False
        return True

    @staticmethod
    def options():
        return list(map(lambda c: c.value, CustomerJobTypes))


def jobs_status_overview(jobs: List[AsyncJob]) -> Dict[str, Any]:
    jobs_by_type = defaultdict(list)
    for job in jobs:
        jobs_by_type[job.job_type].append(job)

    jobs_status = {}
    for job_type, job_collection in jobs_by_type.items():
        overview = defaultdict(int)  # type: Dict[str, int]
        for job in job_collection:
            overview[job.job_last_known_status] += 1
        jobs_status[job_type] = dict(overview)

    return jobs_status



import numpy as np


def compute_average_precision(recall, precision):
    """Compute the average precision, given the recall and precision curves.
    Code originally from https://github.com/rbgirshick/py-faster-rcnn.
    # Arguments
        recall:    The recall curve (list).
        precision: The precision curve (list).
    # Returns
        The average precision as computed in py-faster-rcnn.
    """
    # correct AP calculation
    # first append sentinel values at the end
    mrec = np.concatenate(([0.0], recall, [1.0]))
    mpre = np.concatenate(([0.0], precision, [0.0]))

    # compute the precision envelope
    for i in range(mpre.size - 1, 0, -1):
        mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

    # to calculate area under PR curve, look for points
    # where X axis (recall) changes value
    i = np.where(mrec[1:] != mrec[:-1])[0]

    # and sum (\Delta recall) * prec
    ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
    return ap



"""Shared stateless utility function library for chipping images"""

import io
import json
import os
from itertools import product
from typing import Dict, List

import boto3
import numpy as np
from botocore.exceptions import ClientError
from PIL import Image

from .constants import (
    ANNOTATION_LOCATION_KEY,
    BOX_TYPE,
    GEOMETRY_KEY,
    HEIGHT_KEY,
    IMAGE_LOCATION_KEY,
    LABEL_KEY,
    TYPE_KEY,
    WIDTH_KEY,
    X_KEY,
    Y_KEY,
)


def split_s3_bucket_key(s3_path: str):
    s3_bucket, s3_key = s3_path.split("//", 1)[-1].split("/", 1)
    return s3_bucket, s3_key


def fetch_image(s3_url: str):
    s3_bucket, s3_key = split_s3_bucket_key(s3_url)
    image = Image.open(
        boto3.resource("s3").Bucket(s3_bucket).Object(s3_key).get()["Body"]
    )
    return image


def fetch_chip(ref_id: str):
    """
    Fetches the locations of the image and its corresponding annotations.

    This function checks if the reference ID starts with "s3" to determine if the
    image and annotations are stored on S3, otherwise it checks the local filesystem.
    If the image or annotations do not exist, it returns None for their locations.

    Args:
        ref_id (str): The reference ID for the image and annotations.

    Returns:
        A tuple containing the location of the image and the annotations.
        If either is not found, None is returned in their place.
    """
    image_loc = None
    annotation_loc = None
    if ref_id.startswith("s3"):
        s3_bucket, s3_key = split_s3_bucket_key(ref_id)
        try:
            boto3.resource("s3").Bucket(s3_bucket).Object(
                s3_key + ".jpeg"
            ).load()
            image_loc = ref_id + ".jpeg"
        except ClientError:
            return None, None
        try:
            boto3.resource("s3").Bucket(s3_bucket).Object(
                s3_key + ".json"
            ).load()
            annotation_loc = ref_id + ".json"
        except ClientError:
            return image_loc, None
    else:
        if os.path.exists(ref_id + ".jpeg"):
            image_loc = ref_id + ".jpeg"
            if os.path.exists(ref_id + ".json"):
                annotation_loc = ref_id + ".json"
    return image_loc, annotation_loc


def write_chip(
    ref_id: str, image: Image.Image, annotations: List[Dict[str, str]]
):
    if ref_id.startswith("s3"):
        s3_bucket, s3_key = split_s3_bucket_key(ref_id)
        byteio = io.BytesIO()
        image.save(byteio, format="jpeg")
        byteio.seek(0)
        boto3.resource("s3").Bucket(s3_bucket).Object(
            s3_key + ".jpeg"
        ).upload_fileobj(byteio)
        annotation_loc = None
        if len(annotations) > 0:
            boto3.resource("s3").Bucket(s3_bucket).Object(
                s3_key + ".json"
            ).put(
                Body=json.dumps(annotations, ensure_ascii=False).encode(
                    "UTF-8"
                ),
                ContentType="application/json",
            )
            annotation_loc = ref_id + ".json"
        return ref_id + ".jpeg", annotation_loc
    else:
        dirs = ref_id.rsplit("/", 1)[0]
        os.makedirs(dirs, exist_ok=True)
        image_loc = ref_id + ".jpeg"
        annotation_loc = None
        image.save(image_loc)
        if len(annotations) > 0:
            annotation_loc = ref_id + ".json"
            with open(annotation_loc, "w", encoding="utf-8") as f:
                json.dump(annotations, f, ensure_ascii=False)
        return image_loc, annotation_loc


def generate_offsets(w: int, h: int, chip_size: int, stride_size: int):
    xs = np.arange(0, w - stride_size, stride_size)
    ys = np.arange(0, h - stride_size, stride_size)
    if len(xs) > 1:
        xs = np.round(xs * (w - chip_size) / xs[-1]).astype(int)
    if len(ys) > 1:
        ys = np.round(ys * (h - chip_size) / ys[-1]).astype(int)
    yield from product(xs, ys)


def chip_annotations(data, x0: int, y0: int, x1: int, y1: int):
    """
    Adjusts the annotations to fit within the chip defined by the rectangle
    with top-left corner (x0, y0) and bottom-right corner (x1, y1).

    Parameters:
        data: List of annotation dictionaries to be adjusted.
        x0: The x-coordinate of the top-left corner of the chip.
        y0: The y-coordinate of the top-left corner of the chip.
        x1: The x-coordinate of the bottom-right corner of the chip.
        y1: The y-coordinate of the bottom-right corner of the chip.

    Returns:
        A list of adjusted annotation dictionaries that fit within the chip.
    """
    annotations = []
    X_1_KEY = "x1"
    Y_1_KEY = "y1"
    for annotation in data:
        geometry = annotation[GEOMETRY_KEY].copy()
        geometry[X_1_KEY] = geometry[X_KEY] + geometry[WIDTH_KEY]
        geometry[Y_1_KEY] = geometry[Y_KEY] + geometry[HEIGHT_KEY]
        geometry[X_KEY] = max(min(geometry[X_KEY], x1), x0) - x0
        geometry[X_1_KEY] = max(min(geometry[X_1_KEY], x1), x0) - x0
        geometry[Y_KEY] = max(min(geometry[Y_KEY], y1), y0) - y0
        geometry[Y_1_KEY] = max(min(geometry[Y_1_KEY], y1), y0) - y0
        geometry[WIDTH_KEY] = geometry[X_1_KEY] - geometry[X_KEY]
        geometry[HEIGHT_KEY] = geometry[Y_1_KEY] - geometry[Y_KEY]
        geometry["area"] = geometry[WIDTH_KEY] * geometry[HEIGHT_KEY]
        if geometry["area"] > 0:
            annotations.append(
                {
                    LABEL_KEY: annotation[LABEL_KEY],
                    TYPE_KEY: BOX_TYPE,
                    GEOMETRY_KEY: {
                        X_KEY: geometry[X_KEY],
                        Y_KEY: geometry[Y_KEY],
                        WIDTH_KEY: geometry[WIDTH_KEY],
                        HEIGHT_KEY: geometry[HEIGHT_KEY],
                    },
                }
            )
    return annotations


def process_chip(chip_arg):
    (
        offset,
        chip_size,
        w,
        h,
        item_ref_id,
        cache_directory,
        image,
        annotations,
    ) = chip_arg
    x0, y0 = map(int, offset)
    x1 = min(x0 + chip_size, w)
    y1 = min(y0 + chip_size, h)
    ref_id = f"{cache_directory}/{item_ref_id}_{x0}_{y0}_{x1}_{y1}"
    chipped_image_loc, chipped_annotation_loc = fetch_chip(ref_id)
    if chipped_image_loc:
        return {
            IMAGE_LOCATION_KEY: chipped_image_loc,
            ANNOTATION_LOCATION_KEY: chipped_annotation_loc,
        }
    chipped_image = image.crop((x0, y0, x1, y1))
    chipped_annotations = chip_annotations(annotations, x0, y0, x1, y1)
    chipped_image_loc, chipped_annotation_loc = write_chip(
        ref_id, chipped_image, chipped_annotations
    )
    return {
        IMAGE_LOCATION_KEY: chipped_image_loc,
        ANNOTATION_LOCATION_KEY: chipped_annotation_loc,
    }



from typing import Set

from .constants import (
    DATASET_ID_KEY,
    ERROR_CODES,
    ERROR_ITEMS,
    ERROR_PAYLOAD,
    IGNORED_ITEMS,
    NEW_ITEMS,
    UPDATED_ITEMS,
)
from .dataset_item import DatasetItem


def json_list_to_dataset_item(item_list):
    return [DatasetItem.from_json(item) for item in item_list]


class UploadResponse:
    """Response for long upload job. For internal use only!

    Parameters:
        json: Payload from which to construct the UploadResponse.

    Attributes:
        dataset_id: The scale-generated id for the dataset that was uploaded to
        new_items: How many items are new in the upload
        updated_items: How many items were updated
        ignored_items: How many items were ignored
        upload_errors: A list of errors encountered during upload
        error_codes: A set of all the error codes encountered during upload
        error_payload: The detailed error payload returned from the endpoint.
    """

    def __init__(self, json: dict):
        dataset_id = json.get(DATASET_ID_KEY)
        new_items = json.get(NEW_ITEMS, 0)
        updated_items = json.get(UPDATED_ITEMS, 0)
        ignored_items = json.get(IGNORED_ITEMS, 0)
        upload_errors = json.get(ERROR_ITEMS, 0)
        upload_error_payload = json_list_to_dataset_item(
            json.get(ERROR_PAYLOAD, [])
        )

        self.dataset_id = dataset_id
        self.new_items = new_items
        self.updated_items = updated_items
        self.ignored_items = ignored_items
        self.upload_errors = upload_errors
        self.error_codes: Set[str] = set()
        self.error_payload = upload_error_payload

    def __repr__(self):
        return f"UploadResponse(json={self.json()})"

    def __eq__(self, other):
        return self.json() == other.json()

    def update_response(self, json):
        """
        :param json: { new_items: int, updated_items: int, ignored_items: int, upload_errors: int, }
        """
        assert self.dataset_id == json.get(DATASET_ID_KEY)
        self.new_items += json.get(NEW_ITEMS, 0)
        self.updated_items += json.get(UPDATED_ITEMS, 0)
        self.ignored_items += json.get(IGNORED_ITEMS, 0)
        self.upload_errors += json.get(ERROR_ITEMS, 0)
        if json.get(ERROR_PAYLOAD, None):
            self.error_payload.extend(json.get(ERROR_PAYLOAD, None))

    def record_error(self, response, num_uploads):
        """
        :param response: json response
        :param num_uploads: len of itemss tried to upload
        """
        status = response.status_code
        self.error_codes.add(status)
        self.upload_errors += num_uploads

    def json(self) -> dict:
        """
        return: { new_items: int, updated_items: int, ignored_items: int, upload_errors: int, }
        """
        result = {
            DATASET_ID_KEY: self.dataset_id,
            NEW_ITEMS: self.new_items,
            UPDATED_ITEMS: self.updated_items,
            IGNORED_ITEMS: self.ignored_items,
            ERROR_ITEMS: self.upload_errors,
        }
        if self.error_payload:
            result[ERROR_PAYLOAD] = self.error_payload

        if self.error_codes:
            result[ERROR_ITEMS] = self.upload_errors
            result[ERROR_CODES] = self.error_codes
        return result



from typing import List

from nucleus.async_job import AsyncJob
from nucleus.connection import Connection

from .constants import EVAL_FUNCTION_KEY, SCENARIO_TEST_ID_KEY, EntityLevel
from .data_transfer_objects.eval_function import (
    CreateEvalFunction,
    EvalFunctionEntry,
    GetEvalFunctions,
)
from .data_transfer_objects.scenario_test import (
    CreateScenarioTestRequest,
    EvalFunctionListEntry,
)
from .errors import CreateScenarioTestError
from .eval_functions.available_eval_functions import AvailableEvalFunctions
from .eval_functions.base_eval_function import EvalFunctionConfig
from .scenario_test import ScenarioTest

SUCCESS_KEY = "success"
EVAL_FUNCTIONS_KEY = "eval_functions"


class Validate:
    """Model CI Python Client extension."""

    def __init__(self, api_key: str, endpoint: str):
        self.connection = Connection(api_key, endpoint)

    def __repr__(self):
        return f"Validate(connection='{self.connection}')"

    def __eq__(self, other):
        return self.connection == other.connection

    @property
    def eval_functions(self) -> AvailableEvalFunctions:
        """List all available evaluation functions which can be used to set up evaluation criteria.::

            import nucleus
            client = nucleus.NucleusClient("YOUR_SCALE_API_KEY")

            # Creates an EvaluationCriterion by using a comparison op
            scenario_test_criterion = client.validate.eval_functions.bbox_iou() > 0.5

        Returns:
            :class:`AvailableEvalFunctions`: A container for all the available eval functions
        """
        response = self.connection.get(
            "validate/eval_fn",
        )
        payload = GetEvalFunctions.parse_obj(response)
        return AvailableEvalFunctions(payload.eval_functions)

    def create_scenario_test(
        self,
        name: str,
        slice_id: str,
        evaluation_functions: List[EvalFunctionConfig],
    ) -> ScenarioTest:
        """Creates a new Scenario Test from an existing Nucleus :class:`Slice`:. ::

            import nucleus
            client = nucleus.NucleusClient("YOUR_SCALE_API_KEY")

            scenario_test = client.validate.create_scenario_test(
                name="sample_scenario_test",
                slice_id="YOUR_SLICE_ID",
                evaluation_functions=[client.validate.eval_functions.bbox_iou()]
            )

        Args:
            name: unique name of test
            slice_id: id of (pre-defined) slice of items to evaluate test on.
            evaluation_functions: :class:`EvalFunctionEntry` defines an evaluation metric for the test.
            Created with an element from the list of available eval functions. See :class:`eval_functions`.

        Returns:
            Created ScenarioTest object.
        """
        if not evaluation_functions:
            raise CreateScenarioTestError(
                "Must pass an evaluation_function to the scenario test! I.e. "
                "evaluation_functions=[client.validate.eval_functions.bbox_iou()]"
            )

        external_fns = [
            f.eval_func_entry.is_external_function
            for f in evaluation_functions
        ]
        if any(external_fns):
            assert all(
                external_fns
            ), "Cannot create scenario tests with mixed placeholder and non-placeholder evaluation functions"

        response = self.connection.post(
            CreateScenarioTestRequest(
                name=name,
                slice_id=slice_id,
                evaluation_functions=[
                    EvalFunctionListEntry(
                        id=ef.id, eval_func_arguments=ef.eval_func_arguments
                    )
                    for ef in evaluation_functions
                ],
            ).dict(),
            "validate/scenario_test",
        )
        return ScenarioTest.from_id(
            response[SCENARIO_TEST_ID_KEY], self.connection
        )

    def get_scenario_test(self, scenario_test_id: str) -> ScenarioTest:
        response = self.connection.get(
            f"validate/scenario_test/{scenario_test_id}",
        )
        return ScenarioTest.from_id(
            response["unit_test"]["id"], self.connection
        )

    @property
    def scenario_tests(self) -> List[ScenarioTest]:
        """Lists all Scenario Tests of the current user. ::

            import nucleus
            client = nucleus.NucleusClient("YOUR_SCALE_API_KEY")
            scenario_test = client.validate.create_scenario_test(
                "sample_scenario_test", "slc_bx86ea222a6g057x4380"
            )

            client.validate.scenario_tests

        Returns:
            A list of ScenarioTest objects.
        """
        response = self.connection.get(
            "validate/scenario_test/details",
        )
        tests = [
            ScenarioTest.from_response(payload, self.connection)
            for payload in response
        ]
        return tests

    def delete_scenario_test(self, scenario_test_id: str) -> bool:
        """Deletes a Scenario Test. ::

            import nucleus
            client = nucleus.NucleusClient("YOUR_SCALE_API_KEY")
            scenario_test = client.validate.scenario_tests[0]

            success = client.validate.delete_scenario_test(scenario_test.id)

        Args:
            scenario_test_id: unique ID of scenario test

        Returns:
            Whether deletion was successful.
        """
        response = self.connection.delete(
            f"validate/scenario_test/{scenario_test_id}",
        )
        return response[SUCCESS_KEY]

    def evaluate_model_on_scenario_tests(
        self, model_id: str, scenario_test_names: List[str]
    ) -> AsyncJob:
        """Evaluates the given model on the specified Scenario Tests. ::

            import nucleus
            client = nucleus.NucleusClient("YOUR_SCALE_API_KEY")
            model = client.list_models()[0]
            scenario_test = client.validate.create_scenario_test(
                "sample_scenario_test", "slc_bx86ea222a6g057x4380"
            )

            job = client.validate.evaluate_model_on_scenario_tests(
                model_id=model.id,
                scenario_test_names=["sample_scenario_test"],
            )
            job.sleep_until_complete() # Not required. Will block and update on status of the job.

        Args:
            model_id: ID of model to evaluate
            scenario_test_names: list of scenario test names of test to evaluate

        Returns:
            AsyncJob object of evaluation job
        """
        response = self.connection.post(
            {"test_names": scenario_test_names},
            f"validate/{model_id}/evaluate",
        )
        return AsyncJob.from_json(response, self.connection)

    def metrics(self, model_id: str):
        response = self.connection.post(
            {},
            f"validate/{model_id}/metrics",
        )
        jobs = [AsyncJob.from_json(job, self.connection) for job in response]
        return jobs

    def create_external_eval_function(
        self,
        name: str,
        level: EntityLevel = EntityLevel.ITEM,
    ) -> EvalFunctionEntry:
        """Creates a new external evaluation function. This external function can be used to upload evaluation
        results with functions defined and computed by the customer, without having to share the source code of the
        respective function.

        Args:
            name: unique name of evaluation function
            level: level at which the eval function is run, defaults to EntityLevel.ITEM.

        Raises:
            - NucleusAPIError if the creation of the function fails on the server side
            - ValidationError if the evaluation name is not well defined

        Returns:
            Created EvalFunctionConfig object.

        """

        response = self.connection.post(
            CreateEvalFunction(
                name=name,
                is_external_function=True,
                serialized_fn=None,
                raw_source=None,
                level=level,
            ).dict(),
            "validate/eval_fn",
        )
        return EvalFunctionEntry.parse_obj(response[EVAL_FUNCTION_KEY])



from dataclasses import dataclass
from enum import Enum
from typing import Optional

from .constants import (
    DATASET_ITEM_ID_KEY,
    GROUND_TRUTH_ANNOTATION_ID_KEY,
    GROUND_TRUTH_ANNOTATION_LABEL_KEY,
    IOU_KEY,
    MODEL_PREDICTION_ID_KEY,
    MODEL_PREDICTION_LABEL_KEY,
    MODEL_RUN_ID_KEY,
    TRUE_POSITIVE_KEY,
)


class ConfusionCategory(Enum):
    TRUE_POSITIVE = "true_positive"
    FALSE_POSITIVE = "false_positive"
    FALSE_NEGATIVE = "false_negative"


def infer_confusion_category(
    true_positive: bool,
    ground_truth_annotation_label: str,
    model_prediction_label: str,
):
    confusion_category = ConfusionCategory.FALSE_NEGATIVE

    if (
        true_positive
        or model_prediction_label == ground_truth_annotation_label
    ):
        confusion_category = ConfusionCategory.TRUE_POSITIVE
    elif model_prediction_label is not None:
        confusion_category = ConfusionCategory.FALSE_POSITIVE

    return confusion_category


@dataclass
class EvaluationMatch:
    """
    EvaluationMatch is a result from a model run evaluation. It can represent a true positive, false positive,
    or false negative.

    The matching only matches the strongest prediction for each annotation, so if there are multiple predictions
    that overlap a single annotation only the one with the highest overlap metric will be matched.

    The model prediction label and the ground truth annotation label can differ for true positives if there is configured
    an allowed_label_mapping for the model run.

    NOTE: There is no iou thresholding applied to these matches, so it is possible to have a true positive with a low
    iou score. If manually rejecting matches remember that a rejected match produces both a false positive and a false
    negative otherwise you'll skew your aggregates.

    Attributes:
        model_run_id (str): The ID of the model run that produced this match.
        model_prediction_id (str): The ID of the model prediction that was matched. None if the match was a false negative.
        ground_truth_annotation_id (str): The ID of the ground truth annotation that was matched. None if the match was a false positive.
        iou (int): The intersection over union score of the match.
        dataset_item_id (str): The ID of the dataset item that was matched.
        confusion_category (ConfusionCategory): The confusion category of the match.
        model_prediction_label (str): The label of the model prediction that was matched. None if the match was a false negative.
        ground_truth_annotation_label (str): The label of the ground truth annotation that was matched. None if the match was a false positive.
    """

    model_run_id: str
    model_prediction_id: Optional[str]  # field is nullable
    ground_truth_annotation_id: Optional[str]  # field is nullable
    iou: float
    dataset_item_id: str
    confusion_category: ConfusionCategory
    model_prediction_label: Optional[str]  # field is nullable
    ground_truth_annotation_label: Optional[str]  # field is nullable

    @classmethod
    def from_json(cls, payload: dict):
        is_true_positive = payload.get(TRUE_POSITIVE_KEY, False)
        model_prediction_label = payload.get(MODEL_PREDICTION_LABEL_KEY, None)
        ground_truth_annotation_label = payload.get(
            GROUND_TRUTH_ANNOTATION_LABEL_KEY, None
        )

        confusion_category = infer_confusion_category(
            true_positive=is_true_positive,
            ground_truth_annotation_label=ground_truth_annotation_label,
            model_prediction_label=model_prediction_label,
        )

        return cls(
            model_run_id=payload[MODEL_RUN_ID_KEY],
            model_prediction_id=payload.get(MODEL_PREDICTION_ID_KEY, None),
            ground_truth_annotation_id=payload.get(
                GROUND_TRUTH_ANNOTATION_ID_KEY, None
            ),
            iou=payload[IOU_KEY],
            dataset_item_id=payload[DATASET_ITEM_ID_KEY],
            confusion_category=confusion_category,
            model_prediction_label=model_prediction_label,
            ground_truth_annotation_label=ground_truth_annotation_label,
        )



import os
import shutil
import subprocess

import click
from shellingham import detect_shell


def generate_completions(shell, completion_path):
    command = f"_NU_COMPLETE={shell}_source nu > {completion_path}"
    subprocess.run(command, shell=True)
    click.echo(f"Generated completions for {shell}: {completion_path}")


@click.command("install-completion")
def install_completion():
    """Install shell completion script to your rc file"""
    shell, _ = detect_shell()
    os.makedirs("~/.config", exist_ok=True)
    append_to_file = None
    if shell == "zsh":
        rc_path = "~/.zshrc"
        completion_path = "~/.config/nu-cli-completions.zsh"
        append_to_file = f". {completion_path}"
    elif shell == "bash":
        rc_path = "~/.bashrc"
        completion_path = "~/.config/nu-cli-completions.bash"
        append_to_file = f". {completion_path}"
    elif shell == "fish":
        rc_path = "~/.config/fish/completions/foo-bar.fish"
        completion_path = "~/.config/fish/completions/nu.fish"
    else:
        raise RuntimeError(f"Unsupported shell {shell} for completions")

    generate_completions(shell, completion_path)

    if append_to_file is not None:
        rc_path_expanded = os.path.expanduser(rc_path)
        rc_bak = f"{rc_path_expanded}.bak"
        shutil.copy(rc_path_expanded, rc_bak)
        click.echo(f"Backed up {rc_path} to {rc_bak}")
        with open(rc_path_expanded, mode="a") as rc_file:
            rc_file.write("\n")
            rc_file.write("# Shell completion for nu\n")
            rc_file.write(append_to_file)

    click.echo(f"Completion script added to {rc_path}")
    click.echo(f"Don't forget to `source {rc_path}")



import asyncio
import time
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    BinaryIO,
    Callable,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import aiohttp
import nest_asyncio
from tqdm import tqdm

from nucleus.constants import DEFAULT_NETWORK_TIMEOUT_SEC
from nucleus.errors import NucleusAPIError
from nucleus.retry_strategy import RetryStrategy

from .logger import logger

if TYPE_CHECKING:
    from . import NucleusClient


@dataclass
class FileFormField:
    name: str
    filename: str
    value: BinaryIO
    content_type: str


FileFormData = Sequence[FileFormField]

UPLOAD_SEMAPHORE = asyncio.Semaphore(10)


class FormDataContextHandler:
    """A context handler for file form data that handles closing all files in a request.

    Why do I need to wrap my requests in such a funny way?

    1. Form data must be regenerated on each request to avoid errors
        see https://github.com/Rapptz/discord.py/issues/6531
    2. Files must be properly open/closed for each request.
    3. We need to be able to do 1/2 above multiple times so that we can implement retries
        properly.

    Write a function that returns a tuple of form data and file pointers, then pass it to the
    constructor of this class, and this class will handle the rest for you.
    """

    def __init__(
        self,
        form_data_and_file_pointers_fn: Callable[
            ..., Tuple[FileFormData, Sequence[BinaryIO]]
        ],
    ):
        self._form_data_and_file_pointer_fn = form_data_and_file_pointers_fn
        self._file_pointers = None

    def __enter__(self):
        (
            file_form_data,
            self._file_pointers,
        ) = self._form_data_and_file_pointer_fn()
        form = aiohttp.FormData()
        for field in file_form_data:
            form.add_field(
                name=field.name,
                filename=field.filename,
                value=field.value,
                content_type=field.content_type,
            )
        return form

    def __exit__(self, exc_type, exc_val, exc_tb):
        for file_pointer in self._file_pointers:
            file_pointer.close()


def get_event_loop():
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:  # no event loop running:
        loop = asyncio.new_event_loop()
    else:
        nest_asyncio.apply(loop)
    return loop


def make_multiple_requests_concurrently(
    client: "NucleusClient",
    requests: Sequence[Union[FormDataContextHandler, str]],
    route: Optional[str],
    progressbar: tqdm,
):
    """
    Makes an async post request with form data to a Nucleus endpoint.

    Args:
        client: The client to use for the request.
        requests: a list of requests to make. This list either comprises a string of endpoints to request,
        or a list of FormDataContextHandler object which will handle generating form data, and opening/closing files for each request.
        route: A route is required when requests are for Form Data Post requests
        progressbar: A tqdm progress bar to use for showing progress to the user.
    """
    loop = get_event_loop()
    return loop.run_until_complete(
        _request_helper(client, requests, route, progressbar)
    )


async def _request_helper(
    client: "NucleusClient",
    requests: Sequence[Union[FormDataContextHandler, str]],
    route: Optional[str],
    progressbar: tqdm,
):
    """
    Makes an async requests to a Nucleus endpoint.

    Args:
        client: The client to use for the request.
        requests: a list of requests to make. This list either comprises a string of endpoints to request,
        or a list of FormDataContextHandler object which will handle generating form data, and opening/closing files for each request.
        route: route for the request.
    """
    async with aiohttp.ClientSession() as session:
        tasks = []
        for request in requests:
            if isinstance(request, FormDataContextHandler):
                assert (
                    route
                ), "A route must be specified for FormDataContextHandler requests"
                req = asyncio.ensure_future(
                    _post_form_data(
                        client=client,
                        request=request,
                        route=route,
                        session=session,
                        progressbar=progressbar,
                    )
                )
                tasks.append(req)
            else:
                req = asyncio.ensure_future(
                    _make_request(
                        client=client,
                        request=request,
                        session=session,
                        progressbar=progressbar,
                    )
                )
                tasks.append(req)

        return await asyncio.gather(*tasks)


async def _post_form_data(
    client: "NucleusClient",
    request: FormDataContextHandler,
    route: str,
    session: aiohttp.ClientSession,
    progressbar: tqdm,
):
    """
    Makes an async post request with files to a Nucleus endpoint.

    Args:
        client: The client to use for the request.
        request: The request to make (See FormDataContextHandler for more details.)
        route: route for the request.
        session: The session to use for the request.
    """
    endpoint = f"{client.endpoint}/{route}"
    logger.info("Posting to %s", endpoint)

    async with UPLOAD_SEMAPHORE:
        for sleep_time in RetryStrategy.sleep_times() + [-1]:
            with request as form:
                async with session.post(
                    endpoint,
                    data=form,
                    auth=aiohttp.BasicAuth(client.api_key, ""),
                    timeout=DEFAULT_NETWORK_TIMEOUT_SEC,
                ) as response:
                    data = await _parse_async_response(
                        endpoint, session, response, sleep_time
                    )
                    if data is None:
                        continue

                    progressbar.update(1)
                    return data


async def _make_request(
    client: "NucleusClient",
    request: str,
    session: aiohttp.ClientSession,
    progressbar: tqdm,
):
    """
    Makes an async post request with files to a Nucleus endpoint.

    Args:
        client: The client to use for the request.
        request: The request to make (See FormDataContextHandler for more details.)
        route: route for the request.
        session: The session to use for the request.

    Returns:
        A tuple (endpoint request string, response from endpoint)
    """
    endpoint = f"{client.endpoint}/{request}"
    logger.info("GET %s", endpoint)

    async with UPLOAD_SEMAPHORE:
        for sleep_time in RetryStrategy.sleep_times() + [-1]:
            async with session.get(
                endpoint,
                auth=aiohttp.BasicAuth(client.api_key, ""),
                timeout=DEFAULT_NETWORK_TIMEOUT_SEC,
            ) as response:
                data = await _parse_async_response(
                    endpoint, session, response, sleep_time
                )
                if data is None:
                    continue

                progressbar.update(1)
                return (request, data)


async def _parse_async_response(endpoint, session, response, sleep_time):
    logger.info("API request has response code %s", response.status)

    try:
        data = await response.json()
    except aiohttp.client_exceptions.ContentTypeError:
        # In case of 404, the server returns text
        data = await response.text()
    if response.status in RetryStrategy.statuses and sleep_time != -1:
        time.sleep(sleep_time)
        return None

    if response.status == 503:
        raise TimeoutError(
            "The request to upload your max is timing out, please lower local_files_per_upload_request in your api call."
        )

    if not response.ok:
        raise NucleusAPIError(
            endpoint,
            session.get,
            aiohttp_response=(
                response.status,
                response.reason,
                data,
            ),
        )

    return data
